#pragma once

#include <adwaita.h>

G_BEGIN_DECLS

#define ADW_TYPE_DEMO_PAGE_NAVIGATION_VIEW (adw_demo_page_navigation_view_get_type())

G_DECLARE_FINAL_TYPE (AdwDemoPageNavigationView, adw_demo_page_navigation_view, ADW, DEMO_PAGE_NAVIGATION_VIEW, AdwBin)

G_END_DECLS
