#pragma once

#include <adwaita.h>

G_BEGIN_DECLS

#define ADW_TYPE_DEMO_PAGE_MULTI_LAYOUT (adw_demo_page_multi_layout_get_type())

G_DECLARE_FINAL_TYPE (AdwDemoPageMultiLayout, adw_demo_page_multi_layout, ADW, DEMO_PAGE_MULTI_LAYOUT, AdwBin)

G_END_DECLS
