#pragma once

#include <adwaita.h>

G_BEGIN_DECLS

#define ADW_TYPE_DEMO_PAGE_WRAP_BOX (adw_demo_page_wrap_box_get_type())

G_DECLARE_FINAL_TYPE (AdwDemoPageWrapBox, adw_demo_page_wrap_box, ADW, DEMO_PAGE_WRAP_BOX, AdwBin)

G_END_DECLS
