
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "config.h"
#include "adw-enums.h"
#include "adw-accent-color.h"
#include "adw-alert-dialog.h"
#include "adw-animation.h"
#include "adw-banner.h"
#include "adw-breakpoint.h"
#include "adw-dialog.h"
#include "adw-flap.h"
#include "adw-fold-threshold-policy.h"
#include "adw-easing.h"
#include "adw-header-bar.h"
#include "adw-inline-view-switcher.h"
#include "adw-leaflet.h"
#include "adw-length-unit.h"
#include "adw-navigation-direction.h"
#include "adw-sidebar.h"
#include "adw-style-manager.h"
#include "adw-squeezer.h"
#include "adw-tab-view.h"
#include "adw-toast.h"
#include "adw-toolbar-view.h"
#include "adw-view-switcher.h"
#include "adw-wrap-layout.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "adw-accent-color.h" */

GType
adw_accent_color_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_ACCENT_COLOR_BLUE), "ADW_ACCENT_COLOR_BLUE", "blue" },
        { C_ENUM (ADW_ACCENT_COLOR_TEAL), "ADW_ACCENT_COLOR_TEAL", "teal" },
        { C_ENUM (ADW_ACCENT_COLOR_GREEN), "ADW_ACCENT_COLOR_GREEN", "green" },
        { C_ENUM (ADW_ACCENT_COLOR_YELLOW), "ADW_ACCENT_COLOR_YELLOW", "yellow" },
        { C_ENUM (ADW_ACCENT_COLOR_ORANGE), "ADW_ACCENT_COLOR_ORANGE", "orange" },
        { C_ENUM (ADW_ACCENT_COLOR_RED), "ADW_ACCENT_COLOR_RED", "red" },
        { C_ENUM (ADW_ACCENT_COLOR_PINK), "ADW_ACCENT_COLOR_PINK", "pink" },
        { C_ENUM (ADW_ACCENT_COLOR_PURPLE), "ADW_ACCENT_COLOR_PURPLE", "purple" },
        { C_ENUM (ADW_ACCENT_COLOR_SLATE), "ADW_ACCENT_COLOR_SLATE", "slate" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwAccentColor"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-alert-dialog.h" */

GType
adw_response_appearance_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_RESPONSE_DEFAULT), "ADW_RESPONSE_DEFAULT", "default" },
        { C_ENUM (ADW_RESPONSE_SUGGESTED), "ADW_RESPONSE_SUGGESTED", "suggested" },
        { C_ENUM (ADW_RESPONSE_DESTRUCTIVE), "ADW_RESPONSE_DESTRUCTIVE", "destructive" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwResponseAppearance"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-animation.h" */

GType
adw_animation_state_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_ANIMATION_IDLE), "ADW_ANIMATION_IDLE", "idle" },
        { C_ENUM (ADW_ANIMATION_PAUSED), "ADW_ANIMATION_PAUSED", "paused" },
        { C_ENUM (ADW_ANIMATION_PLAYING), "ADW_ANIMATION_PLAYING", "playing" },
        { C_ENUM (ADW_ANIMATION_FINISHED), "ADW_ANIMATION_FINISHED", "finished" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwAnimationState"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-banner.h" */

GType
adw_banner_button_style_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_BANNER_BUTTON_DEFAULT), "ADW_BANNER_BUTTON_DEFAULT", "default" },
        { C_ENUM (ADW_BANNER_BUTTON_SUGGESTED), "ADW_BANNER_BUTTON_SUGGESTED", "suggested" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwBannerButtonStyle"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-breakpoint.h" */

GType
adw_breakpoint_condition_length_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_BREAKPOINT_CONDITION_MIN_WIDTH), "ADW_BREAKPOINT_CONDITION_MIN_WIDTH", "min-width" },
        { C_ENUM (ADW_BREAKPOINT_CONDITION_MAX_WIDTH), "ADW_BREAKPOINT_CONDITION_MAX_WIDTH", "max-width" },
        { C_ENUM (ADW_BREAKPOINT_CONDITION_MIN_HEIGHT), "ADW_BREAKPOINT_CONDITION_MIN_HEIGHT", "min-height" },
        { C_ENUM (ADW_BREAKPOINT_CONDITION_MAX_HEIGHT), "ADW_BREAKPOINT_CONDITION_MAX_HEIGHT", "max-height" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwBreakpointConditionLengthType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
adw_breakpoint_condition_ratio_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_BREAKPOINT_CONDITION_MIN_ASPECT_RATIO), "ADW_BREAKPOINT_CONDITION_MIN_ASPECT_RATIO", "min-aspect-ratio" },
        { C_ENUM (ADW_BREAKPOINT_CONDITION_MAX_ASPECT_RATIO), "ADW_BREAKPOINT_CONDITION_MAX_ASPECT_RATIO", "max-aspect-ratio" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwBreakpointConditionRatioType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-dialog.h" */

GType
adw_dialog_presentation_mode_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_DIALOG_AUTO), "ADW_DIALOG_AUTO", "auto" },
        { C_ENUM (ADW_DIALOG_FLOATING), "ADW_DIALOG_FLOATING", "floating" },
        { C_ENUM (ADW_DIALOG_BOTTOM_SHEET), "ADW_DIALOG_BOTTOM_SHEET", "bottom-sheet" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwDialogPresentationMode"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-easing.h" */

GType
adw_easing_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_LINEAR), "ADW_LINEAR", "linear" },
        { C_ENUM (ADW_EASE_IN_QUAD), "ADW_EASE_IN_QUAD", "ease-in-quad" },
        { C_ENUM (ADW_EASE_OUT_QUAD), "ADW_EASE_OUT_QUAD", "ease-out-quad" },
        { C_ENUM (ADW_EASE_IN_OUT_QUAD), "ADW_EASE_IN_OUT_QUAD", "ease-in-out-quad" },
        { C_ENUM (ADW_EASE_IN_CUBIC), "ADW_EASE_IN_CUBIC", "ease-in-cubic" },
        { C_ENUM (ADW_EASE_OUT_CUBIC), "ADW_EASE_OUT_CUBIC", "ease-out-cubic" },
        { C_ENUM (ADW_EASE_IN_OUT_CUBIC), "ADW_EASE_IN_OUT_CUBIC", "ease-in-out-cubic" },
        { C_ENUM (ADW_EASE_IN_QUART), "ADW_EASE_IN_QUART", "ease-in-quart" },
        { C_ENUM (ADW_EASE_OUT_QUART), "ADW_EASE_OUT_QUART", "ease-out-quart" },
        { C_ENUM (ADW_EASE_IN_OUT_QUART), "ADW_EASE_IN_OUT_QUART", "ease-in-out-quart" },
        { C_ENUM (ADW_EASE_IN_QUINT), "ADW_EASE_IN_QUINT", "ease-in-quint" },
        { C_ENUM (ADW_EASE_OUT_QUINT), "ADW_EASE_OUT_QUINT", "ease-out-quint" },
        { C_ENUM (ADW_EASE_IN_OUT_QUINT), "ADW_EASE_IN_OUT_QUINT", "ease-in-out-quint" },
        { C_ENUM (ADW_EASE_IN_SINE), "ADW_EASE_IN_SINE", "ease-in-sine" },
        { C_ENUM (ADW_EASE_OUT_SINE), "ADW_EASE_OUT_SINE", "ease-out-sine" },
        { C_ENUM (ADW_EASE_IN_OUT_SINE), "ADW_EASE_IN_OUT_SINE", "ease-in-out-sine" },
        { C_ENUM (ADW_EASE_IN_EXPO), "ADW_EASE_IN_EXPO", "ease-in-expo" },
        { C_ENUM (ADW_EASE_OUT_EXPO), "ADW_EASE_OUT_EXPO", "ease-out-expo" },
        { C_ENUM (ADW_EASE_IN_OUT_EXPO), "ADW_EASE_IN_OUT_EXPO", "ease-in-out-expo" },
        { C_ENUM (ADW_EASE_IN_CIRC), "ADW_EASE_IN_CIRC", "ease-in-circ" },
        { C_ENUM (ADW_EASE_OUT_CIRC), "ADW_EASE_OUT_CIRC", "ease-out-circ" },
        { C_ENUM (ADW_EASE_IN_OUT_CIRC), "ADW_EASE_IN_OUT_CIRC", "ease-in-out-circ" },
        { C_ENUM (ADW_EASE_IN_ELASTIC), "ADW_EASE_IN_ELASTIC", "ease-in-elastic" },
        { C_ENUM (ADW_EASE_OUT_ELASTIC), "ADW_EASE_OUT_ELASTIC", "ease-out-elastic" },
        { C_ENUM (ADW_EASE_IN_OUT_ELASTIC), "ADW_EASE_IN_OUT_ELASTIC", "ease-in-out-elastic" },
        { C_ENUM (ADW_EASE_IN_BACK), "ADW_EASE_IN_BACK", "ease-in-back" },
        { C_ENUM (ADW_EASE_OUT_BACK), "ADW_EASE_OUT_BACK", "ease-out-back" },
        { C_ENUM (ADW_EASE_IN_OUT_BACK), "ADW_EASE_IN_OUT_BACK", "ease-in-out-back" },
        { C_ENUM (ADW_EASE_IN_BOUNCE), "ADW_EASE_IN_BOUNCE", "ease-in-bounce" },
        { C_ENUM (ADW_EASE_OUT_BOUNCE), "ADW_EASE_OUT_BOUNCE", "ease-out-bounce" },
        { C_ENUM (ADW_EASE_IN_OUT_BOUNCE), "ADW_EASE_IN_OUT_BOUNCE", "ease-in-out-bounce" },
        { C_ENUM (ADW_EASE), "ADW_EASE", "ease" },
        { C_ENUM (ADW_EASE_IN), "ADW_EASE_IN", "ease-in" },
        { C_ENUM (ADW_EASE_OUT), "ADW_EASE_OUT", "ease-out" },
        { C_ENUM (ADW_EASE_IN_OUT), "ADW_EASE_IN_OUT", "ease-in-out" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwEasing"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-flap.h" */

GType
adw_flap_fold_policy_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_FLAP_FOLD_POLICY_NEVER), "ADW_FLAP_FOLD_POLICY_NEVER", "never" },
        { C_ENUM (ADW_FLAP_FOLD_POLICY_ALWAYS), "ADW_FLAP_FOLD_POLICY_ALWAYS", "always" },
        { C_ENUM (ADW_FLAP_FOLD_POLICY_AUTO), "ADW_FLAP_FOLD_POLICY_AUTO", "auto" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwFlapFoldPolicy"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
adw_flap_transition_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_FLAP_TRANSITION_TYPE_OVER), "ADW_FLAP_TRANSITION_TYPE_OVER", "over" },
        { C_ENUM (ADW_FLAP_TRANSITION_TYPE_UNDER), "ADW_FLAP_TRANSITION_TYPE_UNDER", "under" },
        { C_ENUM (ADW_FLAP_TRANSITION_TYPE_SLIDE), "ADW_FLAP_TRANSITION_TYPE_SLIDE", "slide" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwFlapTransitionType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-fold-threshold-policy.h" */

GType
adw_fold_threshold_policy_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_FOLD_THRESHOLD_POLICY_MINIMUM), "ADW_FOLD_THRESHOLD_POLICY_MINIMUM", "minimum" },
        { C_ENUM (ADW_FOLD_THRESHOLD_POLICY_NATURAL), "ADW_FOLD_THRESHOLD_POLICY_NATURAL", "natural" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwFoldThresholdPolicy"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-header-bar.h" */

GType
adw_centering_policy_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_CENTERING_POLICY_LOOSE), "ADW_CENTERING_POLICY_LOOSE", "loose" },
        { C_ENUM (ADW_CENTERING_POLICY_STRICT), "ADW_CENTERING_POLICY_STRICT", "strict" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwCenteringPolicy"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-inline-view-switcher.h" */

GType
adw_inline_view_switcher_display_mode_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_INLINE_VIEW_SWITCHER_LABELS), "ADW_INLINE_VIEW_SWITCHER_LABELS", "labels" },
        { C_ENUM (ADW_INLINE_VIEW_SWITCHER_ICONS), "ADW_INLINE_VIEW_SWITCHER_ICONS", "icons" },
        { C_ENUM (ADW_INLINE_VIEW_SWITCHER_BOTH), "ADW_INLINE_VIEW_SWITCHER_BOTH", "both" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwInlineViewSwitcherDisplayMode"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-leaflet.h" */

GType
adw_leaflet_transition_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_LEAFLET_TRANSITION_TYPE_OVER), "ADW_LEAFLET_TRANSITION_TYPE_OVER", "over" },
        { C_ENUM (ADW_LEAFLET_TRANSITION_TYPE_UNDER), "ADW_LEAFLET_TRANSITION_TYPE_UNDER", "under" },
        { C_ENUM (ADW_LEAFLET_TRANSITION_TYPE_SLIDE), "ADW_LEAFLET_TRANSITION_TYPE_SLIDE", "slide" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwLeafletTransitionType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-length-unit.h" */

GType
adw_length_unit_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_LENGTH_UNIT_PX), "ADW_LENGTH_UNIT_PX", "px" },
        { C_ENUM (ADW_LENGTH_UNIT_PT), "ADW_LENGTH_UNIT_PT", "pt" },
        { C_ENUM (ADW_LENGTH_UNIT_SP), "ADW_LENGTH_UNIT_SP", "sp" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwLengthUnit"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-navigation-direction.h" */

GType
adw_navigation_direction_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_NAVIGATION_DIRECTION_BACK), "ADW_NAVIGATION_DIRECTION_BACK", "back" },
        { C_ENUM (ADW_NAVIGATION_DIRECTION_FORWARD), "ADW_NAVIGATION_DIRECTION_FORWARD", "forward" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwNavigationDirection"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-sidebar.h" */

GType
adw_sidebar_mode_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_SIDEBAR_MODE_SIDEBAR), "ADW_SIDEBAR_MODE_SIDEBAR", "sidebar" },
        { C_ENUM (ADW_SIDEBAR_MODE_PAGE), "ADW_SIDEBAR_MODE_PAGE", "page" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwSidebarMode"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-squeezer.h" */

GType
adw_squeezer_transition_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_SQUEEZER_TRANSITION_TYPE_NONE), "ADW_SQUEEZER_TRANSITION_TYPE_NONE", "none" },
        { C_ENUM (ADW_SQUEEZER_TRANSITION_TYPE_CROSSFADE), "ADW_SQUEEZER_TRANSITION_TYPE_CROSSFADE", "crossfade" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwSqueezerTransitionType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-style-manager.h" */

GType
adw_color_scheme_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_COLOR_SCHEME_DEFAULT), "ADW_COLOR_SCHEME_DEFAULT", "default" },
        { C_ENUM (ADW_COLOR_SCHEME_FORCE_LIGHT), "ADW_COLOR_SCHEME_FORCE_LIGHT", "force-light" },
        { C_ENUM (ADW_COLOR_SCHEME_PREFER_LIGHT), "ADW_COLOR_SCHEME_PREFER_LIGHT", "prefer-light" },
        { C_ENUM (ADW_COLOR_SCHEME_PREFER_DARK), "ADW_COLOR_SCHEME_PREFER_DARK", "prefer-dark" },
        { C_ENUM (ADW_COLOR_SCHEME_FORCE_DARK), "ADW_COLOR_SCHEME_FORCE_DARK", "force-dark" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwColorScheme"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-tab-view.h" */

GType
adw_tab_view_shortcuts_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (ADW_TAB_VIEW_SHORTCUT_NONE), "ADW_TAB_VIEW_SHORTCUT_NONE", "none" },
        { C_FLAGS (ADW_TAB_VIEW_SHORTCUT_CONTROL_TAB), "ADW_TAB_VIEW_SHORTCUT_CONTROL_TAB", "control-tab" },
        { C_FLAGS (ADW_TAB_VIEW_SHORTCUT_CONTROL_SHIFT_TAB), "ADW_TAB_VIEW_SHORTCUT_CONTROL_SHIFT_TAB", "control-shift-tab" },
        { C_FLAGS (ADW_TAB_VIEW_SHORTCUT_CONTROL_PAGE_UP), "ADW_TAB_VIEW_SHORTCUT_CONTROL_PAGE_UP", "control-page-up" },
        { C_FLAGS (ADW_TAB_VIEW_SHORTCUT_CONTROL_PAGE_DOWN), "ADW_TAB_VIEW_SHORTCUT_CONTROL_PAGE_DOWN", "control-page-down" },
        { C_FLAGS (ADW_TAB_VIEW_SHORTCUT_CONTROL_HOME), "ADW_TAB_VIEW_SHORTCUT_CONTROL_HOME", "control-home" },
        { C_FLAGS (ADW_TAB_VIEW_SHORTCUT_CONTROL_END), "ADW_TAB_VIEW_SHORTCUT_CONTROL_END", "control-end" },
        { C_FLAGS (ADW_TAB_VIEW_SHORTCUT_CONTROL_SHIFT_PAGE_UP), "ADW_TAB_VIEW_SHORTCUT_CONTROL_SHIFT_PAGE_UP", "control-shift-page-up" },
        { C_FLAGS (ADW_TAB_VIEW_SHORTCUT_CONTROL_SHIFT_PAGE_DOWN), "ADW_TAB_VIEW_SHORTCUT_CONTROL_SHIFT_PAGE_DOWN", "control-shift-page-down" },
        { C_FLAGS (ADW_TAB_VIEW_SHORTCUT_CONTROL_SHIFT_HOME), "ADW_TAB_VIEW_SHORTCUT_CONTROL_SHIFT_HOME", "control-shift-home" },
        { C_FLAGS (ADW_TAB_VIEW_SHORTCUT_CONTROL_SHIFT_END), "ADW_TAB_VIEW_SHORTCUT_CONTROL_SHIFT_END", "control-shift-end" },
        { C_FLAGS (ADW_TAB_VIEW_SHORTCUT_ALT_DIGITS), "ADW_TAB_VIEW_SHORTCUT_ALT_DIGITS", "alt-digits" },
        { C_FLAGS (ADW_TAB_VIEW_SHORTCUT_ALT_ZERO), "ADW_TAB_VIEW_SHORTCUT_ALT_ZERO", "alt-zero" },
        { C_FLAGS (ADW_TAB_VIEW_SHORTCUT_ALL_SHORTCUTS), "ADW_TAB_VIEW_SHORTCUT_ALL_SHORTCUTS", "all-shortcuts" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("AdwTabViewShortcuts"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-toast.h" */

GType
adw_toast_priority_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_TOAST_PRIORITY_NORMAL), "ADW_TOAST_PRIORITY_NORMAL", "normal" },
        { C_ENUM (ADW_TOAST_PRIORITY_HIGH), "ADW_TOAST_PRIORITY_HIGH", "high" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwToastPriority"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-toolbar-view.h" */

GType
adw_toolbar_style_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_TOOLBAR_FLAT), "ADW_TOOLBAR_FLAT", "flat" },
        { C_ENUM (ADW_TOOLBAR_RAISED), "ADW_TOOLBAR_RAISED", "raised" },
        { C_ENUM (ADW_TOOLBAR_RAISED_BORDER), "ADW_TOOLBAR_RAISED_BORDER", "raised-border" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwToolbarStyle"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-view-switcher.h" */

GType
adw_view_switcher_policy_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_VIEW_SWITCHER_POLICY_NARROW), "ADW_VIEW_SWITCHER_POLICY_NARROW", "narrow" },
        { C_ENUM (ADW_VIEW_SWITCHER_POLICY_WIDE), "ADW_VIEW_SWITCHER_POLICY_WIDE", "wide" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwViewSwitcherPolicy"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "adw-wrap-layout.h" */

GType
adw_justify_mode_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_JUSTIFY_NONE), "ADW_JUSTIFY_NONE", "none" },
        { C_ENUM (ADW_JUSTIFY_FILL), "ADW_JUSTIFY_FILL", "fill" },
        { C_ENUM (ADW_JUSTIFY_SPREAD), "ADW_JUSTIFY_SPREAD", "spread" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwJustifyMode"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
adw_pack_direction_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_PACK_START_TO_END), "ADW_PACK_START_TO_END", "start-to-end" },
        { C_ENUM (ADW_PACK_END_TO_START), "ADW_PACK_END_TO_START", "end-to-start" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwPackDirection"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
adw_wrap_policy_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ADW_WRAP_MINIMUM), "ADW_WRAP_MINIMUM", "minimum" },
        { C_ENUM (ADW_WRAP_NATURAL), "ADW_WRAP_NATURAL", "natural" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("AdwWrapPolicy"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

