/*
 * Copyright (C) 2019 Alice Mikhaylenko <alicem@gnome.org>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#include "config.h"
#include "adw-navigation-direction.h"

/**
 * AdwNavigationDirection:
 * @ADW_NAVIGATION_DIRECTION_BACK: Corresponds to start or top, depending on orientation and text direction
 * @ADW_NAVIGATION_DIRECTION_FORWARD: Corresponds to end or bottom, depending on orientation and text direction
 *
 * Describes the direction of a swipe navigation gesture.
 */
